// D import file generated from 'gl3n/linalg.d'
module gl3n.linalg;
private 
{
	import std.math : isNaN, isInfinity;
	import std.conv : to;
	import std.traits : isFloatingPoint, isStaticArray, isDynamicArray, isImplicitlyConvertible, isArray;
	import std.string : format, rightJustify;
	import std.array : join;
	import std.algorithm : max, min, reduce;
	import gl3n.math : clamp, PI, sqrt, sin, cos, acos, tan, asin, atan2, almost_equal;
	import gl3n.util : is_vector, is_matrix, is_quaternion, TupleRange;
}
version (NoReciprocalMul)
{
	private enum rmul = false;

}
else
{
	private enum rmul = true;

}
template Vector(type, int dimension_)
{
	struct Vector
	{
		static assert(dimension > 0, "0 dimensional vectors don't exist.");
		alias type vt;
		static const int dimension = dimension_;

		vt[dimension] vector;
		auto @property value_ptr()
		{
			return vector.ptr;
		}

		@property string as_string()
		{
			return format("%s", vector);
		}


		alias as_string toString;
		nothrow pure @safe 
		{
			private ref @property template get_(char coord)
			{
				inout inout(vt) get_()
				{
					return vector[coord_to_index!coord];
				}

			}


			alias get_!'x' x;
			alias x u;
			alias x s;
			alias x r;
			static if (dimension >= 2)
			{
				alias get_!'y' y;
				alias y v;
				alias y t;
				alias y g;
			}
			static if (dimension >= 3)
			{
				alias get_!'z' z;
				alias z b;
				alias z p;
			}
			static if (dimension >= 4)
			{
				alias get_!'w' w;
				alias w a;
				alias w q;
			}
			static if (dimension == 2)
			{
				enum Vector e1 = Vector(1.to!vt, 0.to!vt);
				enum Vector e2 = Vector(0.to!vt, 1.to!vt);
			}
			else
			{
				static if (dimension == 3)
				{
					enum Vector e1 = Vector(1.to!vt, 0.to!vt, 0.to!vt);
					enum Vector e2 = Vector(0.to!vt, 1.to!vt, 0.to!vt);
					enum Vector e3 = Vector(0.to!vt, 0.to!vt, 1.to!vt);
				}
				else
				{
					static if (dimension == 4)
					{
						enum Vector e1 = Vector(1.to!vt, 0.to!vt, 0.to!vt, 0.to!vt);
						enum Vector e2 = Vector(0.to!vt, 1.to!vt, 0.to!vt, 0.to!vt);
						enum Vector e3 = Vector(0.to!vt, 0.to!vt, 1.to!vt, 0.to!vt);
						enum Vector e4 = Vector(0.to!vt, 0.to!vt, 0.to!vt, 1.to!vt);
					}
				}
			}
			static template isCompatibleVectorImpl(int d) if (d <= dimension)
			{
				void isCompatibleVectorImpl(Vector!(vt, d) vec)
				{
				}

			}

			template isCompatibleVector(T)
			{
				enum isCompatibleVector = is(typeof(isCompatibleVectorImpl(T.init)));
			}
			static template isCompatibleMatrixImpl(int r, int c)
			{
				void isCompatibleMatrixImpl(Matrix!(vt, r, c) m)
				{
				}

			}

			template isCompatibleMatrix(T)
			{
				enum isCompatibleMatrix = is(typeof(isCompatibleMatrixImpl(T.init)));
			}
			private template construct(int i, T, Tail...)
			{
				void construct(T head, Tail tail)
				{
					static if (i >= dimension)
					{
						static assert(false, "Too many arguments passed to constructor");
					}
					else
					{
						static if (is(T : vt))
						{
							vector[i] = head;
							construct!(i + 1)(tail);
						}
						else
						{
							static if (isDynamicArray!T)
							{
								static assert(Tail.length == 0 && i == 0, "dynamic array can not be passed together with other arguments");
								vector[] = head[];
							}
							else
							{
								static if (isStaticArray!T)
								{
									vector[i..i + T.length] = head[];
									construct!(i + T.length)(tail);
								}
								else
								{
									static if (isCompatibleVector!T)
									{
										vector[i..i + T.dimension] = head.vector[];
										construct!(i + T.dimension)(tail);
									}
									else
									{
										static assert(false, "Vector constructor argument must be of type " ~ vt.stringof ~ " or Vector, not " ~ T.stringof);
									}

								}

							}

						}

					}

				}

			}

			private template construct(int i)
			{
				void construct()
				{
					static assert(i == dimension, "Not enough arguments passed to constructor");
				}

			}

			template __ctor(Args...)
			{
				this(Args args)
				{
					construct!0(args);
				}

			}
			template __ctor(T) if (is_vector!T && is(T.vt : vt) && T.dimension >= dimension)
			{
				this(T vec)
				{
					foreach (i; TupleRange!(0, dimension))
					{
						vector[i] = vec.vector[i];
					}
				}

			}
			template __ctor()
			{
				this(vt value)
				{
					clear(value);
				}

			}
			@property const bool isFinite()
			{
				foreach (v; vector)
				{
					if (isNaN(v) || isInfinity(v))
					{
						return false;
					}
				}
				return true;
			}


			deprecated("Use isFinite instead of ok") alias isFinite ok;

			void clear(vt value)
			{
				foreach (i; TupleRange!(0, dimension))
				{
					vector[i] = value;
				}
			}

			template coord_to_index(char c)
			{
				static if (c == 'x' || c == 'r' || c == 'u' || c == 's')
				{
					enum coord_to_index = 0;
				}
				else
				{
					static if (c == 'y' || c == 'g' || c == 'v' || c == 't')
					{
						enum coord_to_index = 1;
					}
					else
					{
						static if (c == 'z' || c == 'b' || c == 'p')
						{
							static assert(dimension >= 3, "the " ~ c ~ " property is only available on vectors with a third dimension.");
							enum coord_to_index = 2;
						}
						else
						{
							static if (c == 'w' || c == 'a' || c == 'q')
							{
								static assert(dimension >= 4, "the " ~ c ~ " property is only available on vectors with a fourth dimension.");
								enum coord_to_index = 3;
							}
							else
							{
								static assert(false, "accepted coordinates are x, s, r, u, y, g, t, v, z, p, b, w, q and a not " ~ c ~ ".");
							}
						}
					}
				}
			}
			static if (dimension == 2)
			{
				void set(vt x, vt y)
				{
					vector[0] = x;
					vector[1] = y;
				}

			}
			static if (dimension == 3)
			{
				void set(vt x, vt y, vt z)
				{
					vector[0] = x;
					vector[1] = y;
					vector[2] = z;
				}

			}
			static if (dimension == 4)
			{
				void set(vt x, vt y, vt z, vt w)
				{
					vector[0] = x;
					vector[1] = y;
					vector[2] = z;
					vector[3] = w;
				}

			}
			void update(Vector!(vt, dimension) other)
			{
				vector = other.vector;
			}

			private template dispatchImpl(int i, string s, int size)
			{
				const void dispatchImpl(ref vt[size] result)
				{
					static if (s.length > 0)
					{
						result[i] = vector[coord_to_index!(s[0])];
						dispatchImpl!(i + 1, s[1 .. $])(result);
					}

				}

			}

			@property template opDispatch(string s)
			{
				const Vector!(vt, s.length) opDispatch()
				{
					vt[s.length] ret;
					dispatchImpl!(0, s)(ret);
					Vector!(vt, s.length) ret_vec;
					ret_vec.vector = ret;
					return ret_vec;
				}

			}

			@property const real magnitude_squared()
			{
				real temp = 0;
				foreach (index; TupleRange!(0, dimension))
				{
					temp += vector[index] ^^ 2;
				}
				return temp;
			}


			@property const real magnitude()
			{
				return sqrt(magnitude_squared);
			}


			alias magnitude_squared length_squared;
			alias magnitude length;
			void normalize()
			{
				real len = length;
				if (len != 0)
				{
					foreach (index; TupleRange!(0, dimension))
					{
						vector[index] /= len;
					}
				}
			}

			@property const Vector normalized()
			{
				Vector ret;
				ret.update(this);
				ret.normalize();
				return ret;
			}


			template opUnary(string op : "-")
			{
				const Vector opUnary()
				{
					Vector ret;
					foreach (index; TupleRange!(0, dimension))
					{
						ret.vector[index] = -vector[index];
					}
					return ret;
				}

			}
			template opBinary(string op : "*")
			{
				const Vector opBinary(vt r)
				{
					Vector ret;
					foreach (index; TupleRange!(0, dimension))
					{
						ret.vector[index] = vector[index] * r;
					}
					return ret;
				}

			}
			template opBinary(string op) if (op == "+" || op == "-")
			{
				const Vector opBinary(Vector r)
				{
					Vector ret;
					foreach (index; TupleRange!(0, dimension))
					{
						ret.vector[index] = mixin("vector[index]" ~ op ~ "r.vector[index]");
					}
					return ret;
				}

			}
			template opBinary(string op : "*")
			{
				const vt opBinary(Vector r)
				{
					return dot(this, r);
				}

			}
			template opBinary(string op : "*", T) if (isCompatibleMatrix!T && T.rows == dimension)
			{
				const Vector!(vt, T.cols) opBinary(T inp)
				{
					Vector!(vt, T.cols) ret;
					ret.clear(0);
					foreach (c; TupleRange!(0, T.cols))
					{
						foreach (r; TupleRange!(0, T.rows))
						{
							ret.vector[c] += vector[r] * inp.matrix[r][c];
						}
					}
					return ret;
				}

			}
			template opBinaryRight(string op, T) if (!is_vector!T && !is_matrix!T && !is_quaternion!T)
			{
				auto const opBinaryRight(T inp)
				{
					return this.opBinary!op(inp);
				}

			}
			template opOpAssign(string op : "*")
			{
				void opOpAssign(vt r)
				{
					foreach (index; TupleRange!(0, dimension))
					{
						vector[index] *= r;
					}
				}

			}
			template opOpAssign(string op) if (op == "+" || op == "-")
			{
				void opOpAssign(Vector r)
				{
					foreach (index; TupleRange!(0, dimension))
					{
						mixin("vector[index]" ~ op ~ "= r.vector[index];");
					}
				}

			}
			const const int opCmp(ref const Vector vec)
			{
				foreach (i, a; vector)
				{
					if (a < vec.vector[i])
					{
						return -1;
					}
					else
						if (a > vec.vector[i])
						{
							return 1;
						}
				}
				return 0;
			}


			const template opEquals(T) if (!isArray!T && T.dimension == dimension)
			{
				bool opEquals(const T vec)
				{
					return vector == vec.vector;
				}

			}

			const template opEquals(T) if (!isArray!T && !is_vector!T)
			{
				bool opEquals(const(T)[] array)
				{
					if (array.length != dimension)
					{
						return false;
					}
					foreach (index; TupleRange!(0, dimension))
					{
						if (vector[index] != array[index])
						{
							return false;
						}
					}
					return true;
				}

			}

			template opCast(T : bool)
			{
				const bool opCast()
				{
					return isFinite;
				}

			}
		}
	}
}
nothrow pure @safe template dot(T) if (is_vector!T)
{
	T.vt dot(const T veca, const T vecb)
	{
		T.vt temp = 0;
		foreach (index; TupleRange!(0, T.dimension))
		{
			temp += veca.vector[index] * vecb.vector[index];
		}
		return temp;
	}

}

nothrow pure @safe template cross(T) if (is_vector!T && T.dimension == 3)
{
	T cross(const T veca, const T vecb)
	{
		return T(veca.y * vecb.z - vecb.y * veca.z, veca.z * vecb.x - vecb.z * veca.x, veca.x * vecb.y - vecb.x * veca.y);
	}

}

nothrow pure @safe template distance(T) if (is_vector!T)
{
	T.vt distance(const T veca, const T vecb)
	{
		return (veca - vecb).length;
	}

}

alias Vector!(float, 2) vec2;
alias Vector!(float, 3) vec3;
alias Vector!(float, 4) vec4;
alias Vector!(double, 2) vec2d;
alias Vector!(double, 3) vec3d;
alias Vector!(double, 4) vec4d;
alias Vector!(int, 2) vec2i;
alias Vector!(int, 3) vec3i;
alias Vector!(int, 4) vec4i;
template Matrix(type, int rows_, int cols_) if (rows_ > 0 && cols_ > 0)
{
	struct Matrix
	{
		alias type mt;
		static const int rows = rows_;

		static const int cols = cols_;

		mt[cols][rows] matrix;
		alias matrix this;
auto @property value_ptr()
		{
			return matrix[0].ptr;
		}

		@property string as_string()
		{
			return format("%s", matrix);
		}


		alias as_string toString;
		@property string as_pretty_string()
		{
			string fmtr = "%s";
			size_t rjust = max(format(fmtr, reduce!max(matrix[])).length, format(fmtr, reduce!min(matrix[])).length) - 1;
			string[] outer_parts;
			foreach (mt[] row; matrix)
			{
				string[] inner_parts;
				foreach (mt col; row)
				{
					inner_parts ~= rightJustify(format(fmtr, col), rjust);
				}
				outer_parts ~= " [" ~ join(inner_parts, ", ") ~ "]";
			}
			return "[" ~ join(outer_parts, "\x0a")[1..$] ~ "]";
		}


		alias as_pretty_string toPrettyString;
		nothrow pure @safe 
		{
			static template isCompatibleMatrixImpl(int r, int c)
			{
				void isCompatibleMatrixImpl(Matrix!(mt, r, c) m)
				{
				}

			}

			template isCompatibleMatrix(T)
			{
				enum isCompatibleMatrix = is(typeof(isCompatibleMatrixImpl(T.init)));
			}
			static template isCompatibleVectorImpl(int d)
			{
				void isCompatibleVectorImpl(Vector!(mt, d) vec)
				{
				}

			}

			template isCompatibleVector(T)
			{
				enum isCompatibleVector = is(typeof(isCompatibleVectorImpl(T.init)));
			}
			private template construct(int i, T, Tail...)
			{
				void construct(T head, Tail tail)
				{
					static if (i >= rows * cols)
					{
						static assert(false, "Too many arguments passed to constructor");
					}
					else
					{
						static if (is(T : mt))
						{
							matrix[i / cols][i % cols] = head;
							construct!(i + 1)(tail);
						}
						else
						{
							static if (is(T == Vector!(mt, cols)))
							{
								static if (i % cols == 0)
								{
									matrix[i / cols] = head.vector;
									construct!(i + T.dimension)(tail);
								}
								else
								{
									static assert(false, "Can't convert Vector into the matrix. Maybe it doesn't align to the columns correctly or dimension doesn't fit");
								}

							}
							else
							{
								static assert(false, "Matrix constructor argument must be of type " ~ mt.stringof ~ " or Vector, not " ~ T.stringof);
							}

						}

					}

				}

			}

			private template construct(int i)
			{
				void construct()
				{
					static assert(i == rows * cols, "Not enough arguments passed to constructor");
				}

			}

			template __ctor(Args...)
			{
				this(Args args)
				{
					construct!0(args);
				}

			}
			template __ctor(T) if (is_matrix!T && T.cols >= cols && T.rows >= rows)
			{
				this(T mat)
				{
					foreach (r; TupleRange!(0, rows))
					{
						foreach (c; TupleRange!(0, cols))
						{
							matrix[r][c] = mat.matrix[r][c];
						}
					}
				}

			}
			template __ctor(T) if (is_matrix!T && T.cols < cols && T.rows < rows)
			{
				this(T mat)
				{
					make_identity();
					foreach (r; TupleRange!(0, T.rows))
					{
						foreach (c; TupleRange!(0, T.cols))
						{
							matrix[r][c] = mat.matrix[r][c];
						}
					}
				}

			}
			template __ctor()
			{
				this(mt value)
				{
					clear(value);
				}

			}
			@property const bool isFinite()
			{
				foreach (row; matrix)
				{
					foreach (col; row)
					{
						if (isNaN(col) || isInfinity(col))
						{
							return false;
						}
					}
				}
				return true;
			}


			deprecated("Use isFinite instead of ok") alias isFinite ok;

			void clear(mt value)
			{
				foreach (r; TupleRange!(0, rows))
				{
					foreach (c; TupleRange!(0, cols))
					{
						matrix[r][c] = value;
					}
				}
			}

			static if (rows == cols)
			{
				void make_identity()
				{
					clear(0);
					foreach (r; TupleRange!(0, rows))
					{
						matrix[r][r] = 1;
					}
				}

				static @property Matrix identity()
				{
					Matrix ret;
					ret.clear(0);
					foreach (r; TupleRange!(0, rows))
					{
						ret.matrix[r][r] = 1;
					}
					return ret;
				}


				void transpose()
				{
					matrix = transposed().matrix;
				}

			}
			@property const Matrix!(mt, cols, rows) transposed()
			{
				typeof(return) ret;
				foreach (r; TupleRange!(0, rows))
				{
					foreach (c; TupleRange!(0, cols))
					{
						ret.matrix[c][r] = matrix[r][c];
					}
				}
				return ret;
			}


			static if (rows == 2 && cols == 2)
			{
				@property const mt det()
				{
					return matrix[0][0] * matrix[1][1] - matrix[0][1] * matrix[1][0];
				}


				private const Matrix invert(ref Matrix mat)
				{
					static if (isFloatingPoint!mt && rmul)
					{
						mt d = 1 / det;
						mat.matrix = [[matrix[1][1] * d, -matrix[0][1] * d], [-matrix[1][0] * d, matrix[0][0] * d]];
					}
					else
					{
						mt d = det;
						mat.matrix = [[matrix[1][1] / d, -matrix[0][1] / d], [-matrix[1][0] / d, matrix[0][0] / d]];
					}

					return mat;
				}


				static Matrix scaling(mt x, mt y)
				{
					Matrix ret = Matrix.identity;
					ret.matrix[0][0] = x;
					ret.matrix[1][1] = y;
					return ret;
				}


				Matrix scale(mt x, mt y)
				{
					this = Matrix.scaling(x, y) * this;
					return this;
				}

			}
			else
			{
				static if (rows == 3 && cols == 3)
				{
					@property const mt det()
					{
						return matrix[0][0] * matrix[1][1] * matrix[2][2] + matrix[0][1] * matrix[1][2] * matrix[2][0] + matrix[0][2] * matrix[1][0] * matrix[2][1] - matrix[0][2] * matrix[1][1] * matrix[2][0] - matrix[0][1] * matrix[1][0] * matrix[2][2] - matrix[0][0] * matrix[1][2] * matrix[2][1];
					}


					private const Matrix invert(ref Matrix mat)
					{
						static if (isFloatingPoint!mt && rmul)
						{
							mt d = 1 / det;
							enum op = "*";
						}
						else
						{
							mt d = det;
							enum op = "/";
						}

						mixin("\x0a            mat.matrix = [[(matrix[1][1] * matrix[2][2] - matrix[1][2] * matrix[2][1])" ~ op ~ "d,\x0a                           (matrix[0][2] * matrix[2][1] - matrix[0][1] * matrix[2][2])" ~ op ~ "d,\x0a                           (matrix[0][1] * matrix[1][2] - matrix[0][2] * matrix[1][1])" ~ op ~ "d],\x0a                          [(matrix[1][2] * matrix[2][0] - matrix[1][0] * matrix[2][2])" ~ op ~ "d,\x0a                           (matrix[0][0] * matrix[2][2] - matrix[0][2] * matrix[2][0])" ~ op ~ "d,\x0a                           (matrix[0][2] * matrix[1][0] - matrix[0][0] * matrix[1][2])" ~ op ~ "d],\x0a                          [(matrix[1][0] * matrix[2][1] - matrix[1][1] * matrix[2][0])" ~ op ~ "d,\x0a                           (matrix[0][1] * matrix[2][0] - matrix[0][0] * matrix[2][1])" ~ op ~ "d,\x0a                           (matrix[0][0] * matrix[1][1] - matrix[0][1] * matrix[1][0])" ~ op ~ "d]];\x0a            ");
						return mat;
					}


				}
				else
				{
					static if (rows == 4 && cols == 4)
					{
						@property const mt det()
						{
							return matrix[0][3] * matrix[1][2] * matrix[2][1] * matrix[3][0] - matrix[0][2] * matrix[1][3] * matrix[2][1] * matrix[3][0] - matrix[0][3] * matrix[1][1] * matrix[2][2] * matrix[3][0] + matrix[0][1] * matrix[1][3] * matrix[2][2] * matrix[3][0] + matrix[0][2] * matrix[1][1] * matrix[2][3] * matrix[3][0] - matrix[0][1] * matrix[1][2] * matrix[2][3] * matrix[3][0] - matrix[0][3] * matrix[1][2] * matrix[2][0] * matrix[3][1] + matrix[0][2] * matrix[1][3] * matrix[2][0] * matrix[3][1] + matrix[0][3] * matrix[1][0] * matrix[2][2] * matrix[3][1] - matrix[0][0] * matrix[1][3] * matrix[2][2] * matrix[3][1] - matrix[0][2] * matrix[1][0] * matrix[2][3] * matrix[3][1] + matrix[0][0] * matrix[1][2] * matrix[2][3] * matrix[3][1] + matrix[0][3] * matrix[1][1] * matrix[2][0] * matrix[3][2] - matrix[0][1] * matrix[1][3] * matrix[2][0] * matrix[3][2] - matrix[0][3] * matrix[1][0] * matrix[2][1] * matrix[3][2] + matrix[0][0] * matrix[1][3] * matrix[2][1] * matrix[3][2] + matrix[0][1] * matrix[1][0] * matrix[2][3] * matrix[3][2] - matrix[0][0] * matrix[1][1] * matrix[2][3] * matrix[3][2] - matrix[0][2] * matrix[1][1] * matrix[2][0] * matrix[3][3] + matrix[0][1] * matrix[1][2] * matrix[2][0] * matrix[3][3] + matrix[0][2] * matrix[1][0] * matrix[2][1] * matrix[3][3] - matrix[0][0] * matrix[1][2] * matrix[2][1] * matrix[3][3] - matrix[0][1] * matrix[1][0] * matrix[2][2] * matrix[3][3] + matrix[0][0] * matrix[1][1] * matrix[2][2] * matrix[3][3];
						}


						private const Matrix invert(ref Matrix mat)
						{
							static if (isFloatingPoint!mt && rmul)
							{
								mt d = 1 / det;
								enum op = "*";
							}
							else
							{
								mt d = det;
								enum op = "/";
							}

							mixin("\x0a            mat.matrix = [[(matrix[1][1] * matrix[2][2] * matrix[3][3] + matrix[1][2] * matrix[2][3] * matrix[3][1] + matrix[1][3] * matrix[2][1] * matrix[3][2]\x0a                          - matrix[1][1] * matrix[2][3] * matrix[3][2] - matrix[1][2] * matrix[2][1] * matrix[3][3] - matrix[1][3] * matrix[2][2] * matrix[3][1])" ~ op ~ "d,\x0a                           (matrix[0][1] * matrix[2][3] * matrix[3][2] + matrix[0][2] * matrix[2][1] * matrix[3][3] + matrix[0][3] * matrix[2][2] * matrix[3][1]\x0a                          - matrix[0][1] * matrix[2][2] * matrix[3][3] - matrix[0][2] * matrix[2][3] * matrix[3][1] - matrix[0][3] * matrix[2][1] * matrix[3][2])" ~ op ~ "d,\x0a                           (matrix[0][1] * matrix[1][2] * matrix[3][3] + matrix[0][2] * matrix[1][3] * matrix[3][1] + matrix[0][3] * matrix[1][1] * matrix[3][2]\x0a                          - matrix[0][1] * matrix[1][3] * matrix[3][2] - matrix[0][2] * matrix[1][1] * matrix[3][3] - matrix[0][3] * matrix[1][2] * matrix[3][1])" ~ op ~ "d,\x0a                           (matrix[0][1] * matrix[1][3] * matrix[2][2] + matrix[0][2] * matrix[1][1] * matrix[2][3] + matrix[0][3] * matrix[1][2] * matrix[2][1]\x0a                          - matrix[0][1] * matrix[1][2] * matrix[2][3] - matrix[0][2] * matrix[1][3] * matrix[2][1] - matrix[0][3] * matrix[1][1] * matrix[2][2])" ~ op ~ "d],\x0a                          [(matrix[1][0] * matrix[2][3] * matrix[3][2] + matrix[1][2] * matrix[2][0] * matrix[3][3] + matrix[1][3] * matrix[2][2] * matrix[3][0]\x0a                          - matrix[1][0] * matrix[2][2] * matrix[3][3] - matrix[1][2] * matrix[2][3] * matrix[3][0] - matrix[1][3] * matrix[2][0] * matrix[3][2])" ~ op ~ "d,\x0a                           (matrix[0][0] * matrix[2][2] * matrix[3][3] + matrix[0][2] * matrix[2][3] * matrix[3][0] + matrix[0][3] * matrix[2][0] * matrix[3][2]\x0a                          - matrix[0][0] * matrix[2][3] * matrix[3][2] - matrix[0][2] * matrix[2][0] * matrix[3][3] - matrix[0][3] * matrix[2][2] * matrix[3][0])" ~ op ~ "d,\x0a                           (matrix[0][0] * matrix[1][3] * matrix[3][2] + matrix[0][2] * matrix[1][0] * matrix[3][3] + matrix[0][3] * matrix[1][2] * matrix[3][0]\x0a                          - matrix[0][0] * matrix[1][2] * matrix[3][3] - matrix[0][2] * matrix[1][3] * matrix[3][0] - matrix[0][3] * matrix[1][0] * matrix[3][2])" ~ op ~ "d,\x0a                           (matrix[0][0] * matrix[1][2] * matrix[2][3] + matrix[0][2] * matrix[1][3] * matrix[2][0] + matrix[0][3] * matrix[1][0] * matrix[2][2]\x0a                          - matrix[0][0] * matrix[1][3] * matrix[2][2] - matrix[0][2] * matrix[1][0] * matrix[2][3] - matrix[0][3] * matrix[1][2] * matrix[2][0])" ~ op ~ "d],\x0a                          [(matrix[1][0] * matrix[2][1] * matrix[3][3] + matrix[1][1] * matrix[2][3] * matrix[3][0] + matrix[1][3] * matrix[2][0] * matrix[3][1]\x0a                          - matrix[1][0] * matrix[2][3] * matrix[3][1] - matrix[1][1] * matrix[2][0] * matrix[3][3] - matrix[1][3] * matrix[2][1] * matrix[3][0])" ~ op ~ "d,\x0a                           (matrix[0][0] * matrix[2][3] * matrix[3][1] + matrix[0][1] * matrix[2][0] * matrix[3][3] + matrix[0][3] * matrix[2][1] * matrix[3][0]\x0a                          - matrix[0][0] * matrix[2][1] * matrix[3][3] - matrix[0][1] * matrix[2][3] * matrix[3][0] - matrix[0][3] * matrix[2][0] * matrix[3][1])" ~ op ~ "d,\x0a                           (matrix[0][0] * matrix[1][1] * matrix[3][3] + matrix[0][1] * matrix[1][3] * matrix[3][0] + matrix[0][3] * matrix[1][0] * matrix[3][1]\x0a                          - matrix[0][0] * matrix[1][3] * matrix[3][1] - matrix[0][1] * matrix[1][0] * matrix[3][3] - matrix[0][3] * matrix[1][1] * matrix[3][0])" ~ op ~ "d,\x0a                           (matrix[0][0] * matrix[1][3] * matrix[2][1] + matrix[0][1] * matrix[1][0] * matrix[2][3] + matrix[0][3] * matrix[1][1] * matrix[2][0]\x0a                          - matrix[0][0] * matrix[1][1] * matrix[2][3] - matrix[0][1] * matrix[1][3] * matrix[2][0] - matrix[0][3] * matrix[1][0] * matrix[2][1])" ~ op ~ "d],\x0a                          [(matrix[1][0] * matrix[2][2] * matrix[3][1] + matrix[1][1] * matrix[2][0] * matrix[3][2] + matrix[1][2] * matrix[2][1] * matrix[3][0]\x0a                          - matrix[1][0] * matrix[2][1] * matrix[3][2] - matrix[1][1] * matrix[2][2] * matrix[3][0] - matrix[1][2] * matrix[2][0] * matrix[3][1])" ~ op ~ "d,\x0a                           (matrix[0][0] * matrix[2][1] * matrix[3][2] + matrix[0][1] * matrix[2][2] * matrix[3][0] + matrix[0][2] * matrix[2][0] * matrix[3][1]\x0a                          - matrix[0][0] * matrix[2][2] * matrix[3][1] - matrix[0][1] * matrix[2][0] * matrix[3][2] - matrix[0][2] * matrix[2][1] * matrix[3][0])" ~ op ~ "d,\x0a                           (matrix[0][0] * matrix[1][2] * matrix[3][1] + matrix[0][1] * matrix[1][0] * matrix[3][2] + matrix[0][2] * matrix[1][1] * matrix[3][0]\x0a                          - matrix[0][0] * matrix[1][1] * matrix[3][2] - matrix[0][1] * matrix[1][2] * matrix[3][0] - matrix[0][2] * matrix[1][0] * matrix[3][1])" ~ op ~ "d,\x0a                           (matrix[0][0] * matrix[1][1] * matrix[2][2] + matrix[0][1] * matrix[1][2] * matrix[2][0] + matrix[0][2] * matrix[1][0] * matrix[2][1]\x0a                          - matrix[0][0] * matrix[1][2] * matrix[2][1] - matrix[0][1] * matrix[1][0] * matrix[2][2] - matrix[0][2] * matrix[1][1] * matrix[2][0])" ~ op ~ "d]];\x0a            ");
							return mat;
						}


						static if (isFloatingPoint!mt)
						{
							static private mt[6] cperspective(mt width, mt height, mt fov, mt near, mt far)
							in
							{
								assert(height != 0);
							}
							body
							{
								mt aspect = width / height;
								mt top = near * tan(fov * (PI / 360.000));
								mt bottom = -top;
								mt right = top * aspect;
								mt left = -right;
								return [left, right, bottom, top, near, far];
							}



							static Matrix perspective(mt width, mt height, mt fov, mt near, mt far)
							{
								mt[6] cdata = cperspective(width, height, fov, near, far);
								return perspective(cdata[0], cdata[1], cdata[2], cdata[3], cdata[4], cdata[5]);
							}


							static Matrix perspective(mt left, mt right, mt bottom, mt top, mt near, mt far)
							in
							{
								assert(right - left != 0);
								assert(top - bottom != 0);
								assert(far - near != 0);
							}
							body
							{
								Matrix ret;
								ret.clear(0);
								ret.matrix[0][0] = 2 * near / (right - left);
								ret.matrix[0][2] = (right + left) / (right - left);
								ret.matrix[1][1] = 2 * near / (top - bottom);
								ret.matrix[1][2] = (top + bottom) / (top - bottom);
								ret.matrix[2][2] = -(far + near) / (far - near);
								ret.matrix[2][3] = -(2 * far * near) / (far - near);
								ret.matrix[3][2] = -1;
								return ret;
							}


							static Matrix perspective_inverse(mt width, mt height, mt fov, mt near, mt far)
							{
								mt[6] cdata = cperspective(width, height, fov, near, far);
								return perspective_inverse(cdata[0], cdata[1], cdata[2], cdata[3], cdata[4], cdata[5]);
							}


							static Matrix perspective_inverse(mt left, mt right, mt bottom, mt top, mt near, mt far)
							in
							{
								assert(near != 0);
								assert(far != 0);
							}
							body
							{
								Matrix ret;
								ret.clear(0);
								ret.matrix[0][0] = (right - left) / (2 * near);
								ret.matrix[0][3] = (right + left) / (2 * near);
								ret.matrix[1][1] = (top - bottom) / (2 * near);
								ret.matrix[1][3] = (top + bottom) / (2 * near);
								ret.matrix[2][3] = -1;
								ret.matrix[3][2] = -(far - near) / (2 * far * near);
								ret.matrix[3][3] = (far + near) / (2 * far * near);
								return ret;
							}


							static Matrix orthographic(mt left, mt right, mt bottom, mt top, mt near, mt far)
							in
							{
								assert(right - left != 0);
								assert(top - bottom != 0);
								assert(far - near != 0);
							}
							body
							{
								Matrix ret;
								ret.clear(0);
								ret.matrix[0][0] = 2 / (right - left);
								ret.matrix[0][3] = -(right + left) / (right - left);
								ret.matrix[1][1] = 2 / (top - bottom);
								ret.matrix[1][3] = -(top + bottom) / (top - bottom);
								ret.matrix[2][2] = -2 / (far - near);
								ret.matrix[2][3] = -(far + near) / (far - near);
								ret.matrix[3][3] = 1;
								return ret;
							}


							static Matrix orthographic_inverse(mt left, mt right, mt bottom, mt top, mt near, mt far)
							{
								Matrix ret;
								ret.clear(0);
								ret.matrix[0][0] = (right - left) / 2;
								ret.matrix[0][3] = (right + left) / 2;
								ret.matrix[1][1] = (top - bottom) / 2;
								ret.matrix[1][3] = (top + bottom) / 2;
								ret.matrix[2][2] = (far - near) / -2;
								ret.matrix[2][3] = (far + near) / 2;
								ret.matrix[3][3] = 1;
								return ret;
							}


							static Matrix look_at(Vector!(mt, 3) eye, Vector!(mt, 3) target, Vector!(mt, 3) up)
							{
								alias Vector!(mt, 3) vec3mt;
								vec3mt look_dir = (target - eye).normalized;
								vec3mt up_dir = up.normalized;
								vec3mt right_dir = cross(look_dir, up_dir).normalized;
								vec3mt perp_up_dir = cross(right_dir, look_dir);
								Matrix ret = Matrix.identity;
								ret.matrix[0][0..3] = right_dir.vector[];
								ret.matrix[1][0..3] = perp_up_dir.vector[];
								ret.matrix[2][0..3] = (-look_dir).vector[];
								ret.matrix[0][3] = -dot(eye, right_dir);
								ret.matrix[1][3] = -dot(eye, perp_up_dir);
								ret.matrix[2][3] = dot(eye, look_dir);
								return ret;
							}


						}
					}
				}
			}
			static if (rows == cols && rows >= 3 && rows <= 4)
			{
				static Matrix translation(mt x, mt y, mt z)
				{
					Matrix ret = Matrix.identity;
					ret.matrix[0][cols - 1] = x;
					ret.matrix[1][cols - 1] = y;
					ret.matrix[2][cols - 1] = z;
					return ret;
				}


				Matrix translate(mt x, mt y, mt z)
				{
					this = Matrix.translation(x, y, z) * this;
					return this;
				}

				static Matrix scaling(mt x, mt y, mt z)
				{
					Matrix ret = Matrix.identity;
					ret.matrix[0][0] = x;
					ret.matrix[1][1] = y;
					ret.matrix[2][2] = z;
					return ret;
				}


				Matrix scale(mt x, mt y, mt z)
				{
					this = Matrix.scaling(x, y, z) * this;
					return this;
				}

			}
			static if (rows == cols && rows >= 3)
			{
				static if (isFloatingPoint!mt)
				{
					static Matrix rotation(real alpha, Vector!(mt, 3) axis)
					{
						Matrix mult = Matrix.identity;
						if (axis.length != 1)
						{
							axis.normalize();
						}
						real cosa = cos(alpha);
						real sina = sin(alpha);
						Vector!(mt, 3) temp = (1 - cosa) * axis;
						mult.matrix[0][0] = to!mt(cosa + temp.x * axis.x);
						mult.matrix[0][1] = to!mt(temp.x * axis.y + sina * axis.z);
						mult.matrix[0][2] = to!mt(temp.x * axis.z - sina * axis.y);
						mult.matrix[1][0] = to!mt(temp.y * axis.x - sina * axis.z);
						mult.matrix[1][1] = to!mt(cosa + temp.y * axis.y);
						mult.matrix[1][2] = to!mt(temp.y * axis.z + sina * axis.x);
						mult.matrix[2][0] = to!mt(temp.z * axis.x + sina * axis.y);
						mult.matrix[2][1] = to!mt(temp.z * axis.y - sina * axis.x);
						mult.matrix[2][2] = to!mt(cosa + temp.z * axis.z);
						return mult;
					}


					static Matrix rotation(real alpha, mt x, mt y, mt z)
					{
						return Matrix.rotation(alpha, Vector!(mt, 3)(x, y, z));
					}


					static Matrix xrotation(real alpha)
					{
						Matrix mult = Matrix.identity;
						mt cosamt = to!mt(cos(alpha));
						mt sinamt = to!mt(sin(alpha));
						mult.matrix[1][1] = cosamt;
						mult.matrix[1][2] = -sinamt;
						mult.matrix[2][1] = sinamt;
						mult.matrix[2][2] = cosamt;
						return mult;
					}


					static Matrix yrotation(real alpha)
					{
						Matrix mult = Matrix.identity;
						mt cosamt = to!mt(cos(alpha));
						mt sinamt = to!mt(sin(alpha));
						mult.matrix[0][0] = cosamt;
						mult.matrix[0][2] = sinamt;
						mult.matrix[2][0] = -sinamt;
						mult.matrix[2][2] = cosamt;
						return mult;
					}


					static Matrix zrotation(real alpha)
					{
						Matrix mult = Matrix.identity;
						mt cosamt = to!mt(cos(alpha));
						mt sinamt = to!mt(sin(alpha));
						mult.matrix[0][0] = cosamt;
						mult.matrix[0][1] = -sinamt;
						mult.matrix[1][0] = sinamt;
						mult.matrix[1][1] = cosamt;
						return mult;
					}


					Matrix rotate(real alpha, Vector!(mt, 3) axis)
					{
						this = rotation(alpha, axis) * this;
						return this;
					}

					Matrix rotatex(real alpha)
					{
						this = xrotation(alpha) * this;
						return this;
					}

					Matrix rotatey(real alpha)
					{
						this = yrotation(alpha) * this;
						return this;
					}

					Matrix rotatez(real alpha)
					{
						this = zrotation(alpha) * this;
						return this;
					}

				}
				void set_translation(mt[] values...)
				in
				{
					assert(values.length >= rows - 1);
				}
				body
				{
					foreach (r; TupleRange!(0, rows - 1))
					{
						matrix[r][rows - 1] = values[r];
					}
				}

				void set_translation(Matrix mat)
				{
					foreach (r; TupleRange!(0, rows - 1))
					{
						matrix[r][rows - 1] = mat.matrix[r][rows - 1];
					}
				}

				Matrix get_translation()
				{
					Matrix ret = Matrix.identity;
					foreach (r; TupleRange!(0, rows - 1))
					{
						ret.matrix[r][rows - 1] = matrix[r][rows - 1];
					}
					return ret;
				}

				void set_scale(mt[] values...)
				in
				{
					assert(values.length >= rows - 1);
				}
				body
				{
					foreach (r; TupleRange!(0, rows - 1))
					{
						matrix[r][r] = values[r];
					}
				}

				void set_scale(Matrix mat)
				{
					foreach (r; TupleRange!(0, rows - 1))
					{
						matrix[r][r] = mat.matrix[r][r];
					}
				}

				Matrix get_scale()
				{
					Matrix ret = Matrix.identity;
					foreach (r; TupleRange!(0, rows - 1))
					{
						ret.matrix[r][r] = matrix[r][r];
					}
					return ret;
				}

				void set_rotation(Matrix!(mt, 3, 3) rot)
				{
					foreach (r; TupleRange!(0, 3))
					{
						foreach (c; TupleRange!(0, 3))
						{
							matrix[r][c] = rot[r][c];
						}
					}
				}

				Matrix!(mt, 3, 3) get_rotation()
				{
					Matrix!(mt, 3, 3) ret = Matrix!(mt, 3, 3).identity;
					foreach (r; TupleRange!(0, 3))
					{
						foreach (c; TupleRange!(0, 3))
						{
							ret.matrix[r][c] = matrix[r][c];
						}
					}
					return ret;
				}

			}
			static if (rows == cols && rows >= 2 && rows <= 4)
			{
				@property const Matrix inverse()
				{
					Matrix mat;
					invert(mat);
					return mat;
				}


				void invert()
				{
					Matrix temp;
					invert(temp);
					this.matrix = temp.matrix;
				}

			}
			private const void mms(mt inp, ref Matrix mat)
			{
				for (int r = 0;
				 r < rows; r++)
				{
					{
						for (int c = 0;
						 c < cols; c++)
						{
							{
								mat.matrix[r][c] = matrix[r][c] * inp;
							}
						}
					}
				}
			}


			private template masm(string op)
			{
				const void masm(Matrix inp, ref Matrix mat)
				{
					foreach (r; TupleRange!(0, rows))
					{
						foreach (c; TupleRange!(0, cols))
						{
							mat.matrix[r][c] = mixin("inp.matrix[r][c]" ~ op ~ "matrix[r][c]");
						}
					}
				}

			}

			template opBinary(string op : "*", T) if (isCompatibleMatrix!T && T.rows == cols)
			{
				const Matrix!(mt, rows, T.cols) opBinary(T inp)
				{
					Matrix!(mt, rows, T.cols) ret;
					foreach (r; TupleRange!(0, rows))
					{
						foreach (c; TupleRange!(0, T.cols))
						{
							ret.matrix[r][c] = 0;
							foreach (c2; TupleRange!(0, cols))
							{
								ret.matrix[r][c] += matrix[r][c2] * inp.matrix[c2][c];
							}
						}
					}
					return ret;
				}

			}
			template opBinary(string op : "*", T : Vector!(mt, cols))
			{
				const Vector!(mt, rows) opBinary(T inp)
				{
					Vector!(mt, rows) ret;
					ret.clear(0);
					foreach (c; TupleRange!(0, cols))
					{
						foreach (r; TupleRange!(0, rows))
						{
							ret.vector[r] += matrix[r][c] * inp.vector[c];
						}
					}
					return ret;
				}

			}
			template opBinary(string op : "*")
			{
				const Matrix opBinary(mt inp)
				{
					Matrix ret;
					mms(inp, ret);
					return ret;
				}

			}
			template opBinaryRight(string op : "*")
			{
				const Matrix opBinaryRight(mt inp)
				{
					return this.opBinary!op(inp);
				}

			}
			template opBinary(string op) if (op == "+" || op == "-")
			{
				const Matrix opBinary(Matrix inp)
				{
					Matrix ret;
					masm!op(inp, ret);
					return ret;
				}

			}
			template opOpAssign(string op : "*")
			{
				void opOpAssign(mt inp)
				{
					mms(inp, this);
				}

			}
			template opOpAssign(string op) if (op == "+" || op == "-")
			{
				void opOpAssign(Matrix inp)
				{
					masm!op(inp, this);
				}

			}
			template opCast(T : bool)
			{
				const bool opCast()
				{
					return isFinite;
				}

			}
		}
	}
}
alias Matrix!(float, 2, 2) mat2;
alias Matrix!(float, 3, 3) mat3;
alias Matrix!(float, 3, 4) mat34;
alias Matrix!(float, 4, 4) mat4;
private {}
template Quaternion(type)
{
	struct Quaternion
	{
		alias type qt;
		qt[4] quaternion;
		auto @property value_ptr()
		{
			return quaternion.ptr;
		}

		@property string as_string()
		{
			return format("%s", quaternion);
		}


		alias as_string toString;
		nothrow pure @safe 
		{
			private @property template get_(char coord)
			{
				const qt get_()
				{
					return quaternion[coord_to_index!coord];
				}

			}


			private @property template set_(char coord)
			{
				void set_(qt value)
				{
					quaternion[coord_to_index!coord] = value;
				}

			}


			alias get_!'w' w;
			alias set_!'w' w;
			alias get_!'x' x;
			alias set_!'x' x;
			alias get_!'y' y;
			alias set_!'y' y;
			alias get_!'z' z;
			alias set_!'z' z;
			this(qt w_, qt x_, qt y_, qt z_)
			{
				w = w_;
				x = x_;
				y = y_;
				z = z_;
			}

			this(qt w_, Vector!(qt, 3) vec)
			{
				w = w_;
				quaternion[1..4] = vec.vector[];
			}

			this(Vector!(qt, 4) vec)
			{
				quaternion[] = vec.vector[];
			}

			@property const bool isFinite()
			{
				foreach (q; quaternion)
				{
					if (isNaN(q) || isInfinity(q))
					{
						return false;
					}
				}
				return true;
			}


			deprecated("Use isFinite instead of ok") alias isFinite ok;

			template coord_to_index(char c)
			{
				static if (c == 'w')
				{
					enum coord_to_index = 0;
				}
				else
				{
					static if (c == 'x')
					{
						enum coord_to_index = 1;
					}
					else
					{
						static if (c == 'y')
						{
							enum coord_to_index = 2;
						}
						else
						{
							static if (c == 'z')
							{
								enum coord_to_index = 3;
							}
							else
							{
								static assert(false, "accepted coordinates are x, y, z and w not " ~ c ~ ".");
							}
						}
					}
				}
			}
			@property const real magnitude_squared()
			{
				return to!real(w ^^ 2 + x ^^ 2 + y ^^ 2 + z ^^ 2);
			}


			@property const real magnitude()
			{
				return sqrt(magnitude_squared);
			}


			static @property Quaternion identity()
			{
				return Quaternion(1, 0, 0, 0);
			}


			void make_identity()
			{
				w = 1;
				x = 0;
				y = 0;
				z = 0;
			}

			void invert()
			{
				x = -x;
				y = -y;
				z = -z;
			}

			alias invert conjugate;
			@property const Quaternion inverse()
			{
				return Quaternion(w, -x, -y, -z);
			}


			alias inverse conjugated;
			static Quaternion from_matrix(Matrix!(qt, 3, 3) matrix)
			{
				Quaternion ret;
				auto mat = matrix.matrix;
				qt trace = mat[0][0] + mat[1][1] + mat[2][2];
				if (trace > 0)
				{
					real s = 0.5 / sqrt(trace + 1.00000F);
					ret.w = to!qt(0.25 / s);
					ret.x = to!qt((mat[2][1] - mat[1][2]) * s);
					ret.y = to!qt((mat[0][2] - mat[2][0]) * s);
					ret.z = to!qt((mat[1][0] - mat[0][1]) * s);
				}
				else
					if (mat[0][0] > mat[1][1] && mat[0][0] > mat[2][2])
					{
						real s = 2.00000 * sqrt(1.00000 + mat[0][0] - mat[1][1] - mat[2][2]);
						ret.w = to!qt((mat[2][1] - mat[1][2]) / s);
						ret.x = to!qt(0.25F * s);
						ret.y = to!qt((mat[0][1] + mat[1][0]) / s);
						ret.z = to!qt((mat[0][2] + mat[2][0]) / s);
					}
					else
						if (mat[1][1] > mat[2][2])
						{
							real s = 2.00000 * sqrt(1 + mat[1][1] - mat[0][0] - mat[2][2]);
							ret.w = to!qt((mat[0][2] - mat[2][0]) / s);
							ret.x = to!qt((mat[0][1] + mat[1][0]) / s);
							ret.y = to!qt(0.25F * s);
							ret.z = to!qt((mat[1][2] + mat[2][1]) / s);
						}
						else
						{
							real s = 2.00000 * sqrt(1 + mat[2][2] - mat[0][0] - mat[1][1]);
							ret.w = to!qt((mat[1][0] - mat[0][1]) / s);
							ret.x = to!qt((mat[0][2] + mat[2][0]) / s);
							ret.y = to!qt((mat[1][2] + mat[2][1]) / s);
							ret.z = to!qt(0.25F * s);
						}
				return ret;
			}


			template to_matrix(int rows, int cols) if (rows >= 3 && cols >= 3)
			{
				const Matrix!(qt, rows, cols) to_matrix()
				{
					static if (rows == 3 && cols == 3)
					{
						Matrix!(qt, rows, cols) ret;
					}
					else
					{
						Matrix!(qt, rows, cols) ret = Matrix!(qt, rows, cols).identity;
					}

					qt xx = x ^^ 2;
					qt xy = x * y;
					qt xz = x * z;
					qt xw = x * w;
					qt yy = y ^^ 2;
					qt yz = y * z;
					qt yw = y * w;
					qt zz = z ^^ 2;
					qt zw = z * w;
					ret.matrix[0][0..3] = [1 - 2 * (yy + zz), 2 * (xy - zw), 2 * (xz + yw)];
					ret.matrix[1][0..3] = [2 * (xy + zw), 1 - 2 * (xx + zz), 2 * (yz - xw)];
					ret.matrix[2][0..3] = [2 * (xz - yw), 2 * (yz + xw), 1 - 2 * (xx + yy)];
					return ret;
				}

			}
			void normalize()
			{
				qt m = to!qt(magnitude);
				if (m != 0)
				{
					w = w / m;
					x = x / m;
					y = y / m;
					z = z / m;
				}
			}

			const Quaternion normalized()
			{
				Quaternion ret;
				qt m = to!qt(magnitude);
				if (m != 0)
				{
					ret.w = w / m;
					ret.x = x / m;
					ret.y = y / m;
					ret.z = z / m;
				}
				else
				{
					ret = Quaternion(w, x, y, z);
				}
				return ret;
			}

			@property const real yaw()
			{
				return atan2(to!real(2 * (w * y + x * z)), to!real(w ^^ 2 - x ^^ 2 - y ^^ 2 + z ^^ 2));
			}


			@property const real pitch()
			{
				return asin(to!real(2 * (w * x - y * z)));
			}


			@property const real roll()
			{
				return atan2(to!real(2 * (w * z + x * y)), to!real(w ^^ 2 - x ^^ 2 + y ^^ 2 - z ^^ 2));
			}


			static Quaternion xrotation(real alpha)
			{
				Quaternion ret;
				alpha /= 2;
				ret.w = to!qt(cos(alpha));
				ret.x = to!qt(sin(alpha));
				ret.y = 0;
				ret.z = 0;
				return ret;
			}


			static Quaternion yrotation(real alpha)
			{
				Quaternion ret;
				alpha /= 2;
				ret.w = to!qt(cos(alpha));
				ret.x = 0;
				ret.y = to!qt(sin(alpha));
				ret.z = 0;
				return ret;
			}


			static Quaternion zrotation(real alpha)
			{
				Quaternion ret;
				alpha /= 2;
				ret.w = to!qt(cos(alpha));
				ret.x = 0;
				ret.y = 0;
				ret.z = to!qt(sin(alpha));
				return ret;
			}


			static Quaternion axis_rotation(real alpha, Vector!(qt, 3) axis)
			{
				if (alpha == 0)
				{
					return Quaternion.identity;
				}
				Quaternion ret;
				alpha /= 2;
				qt sinaqt = to!qt(sin(alpha));
				ret.w = to!qt(cos(alpha));
				ret.x = axis.x * sinaqt;
				ret.y = axis.y * sinaqt;
				ret.z = axis.z * sinaqt;
				return ret;
			}


			static Quaternion euler_rotation(real heading, real attitude, real bank)
			{
				Quaternion ret;
				real c1 = cos(heading / 2);
				real s1 = sin(heading / 2);
				real c2 = cos(attitude / 2);
				real s2 = sin(attitude / 2);
				real c3 = cos(bank / 2);
				real s3 = sin(bank / 2);
				ret.w = to!qt(c1 * c2 * c3 - s1 * s2 * s3);
				ret.x = to!qt(s1 * s2 * c3 + c1 * c2 * s3);
				ret.y = to!qt(s1 * c2 * c3 + c1 * s2 * s3);
				ret.z = to!qt(c1 * s2 * c3 - s1 * c2 * s3);
				return ret;
			}


			Quaternion rotatex(real alpha)
			{
				this = xrotation(alpha) * this;
				return this;
			}

			Quaternion rotatey(real alpha)
			{
				this = yrotation(alpha) * this;
				return this;
			}

			Quaternion rotatez(real alpha)
			{
				this = zrotation(alpha) * this;
				return this;
			}

			Quaternion rotate_axis(real alpha, Vector!(qt, 3) axis)
			{
				this = axis_rotation(alpha, axis) * this;
				return this;
			}

			Quaternion rotate_euler(real heading, real attitude, real bank)
			{
				this = euler_rotation(heading, attitude, bank) * this;
				return this;
			}

			template opBinary(string op : "*")
			{
				const Quaternion opBinary(Quaternion inp)
				{
					Quaternion ret;
					ret.w = -x * inp.x - y * inp.y - z * inp.z + w * inp.w;
					ret.x = x * inp.w + y * inp.z - z * inp.y + w * inp.x;
					ret.y = -x * inp.z + y * inp.w + z * inp.x + w * inp.y;
					ret.z = x * inp.y - y * inp.x + z * inp.w + w * inp.z;
					return ret;
				}

			}
			template opBinaryRight(string op, T) if (!is_quaternion!T)
			{
				auto const opBinaryRight(T inp)
				{
					return this.opBinary!op(inp);
				}

			}
			template opBinary(string op) if (op == "+" || op == "-")
			{
				const Quaternion opBinary(Quaternion inp)
				{
					Quaternion ret;
					mixin("ret.w = w" ~ op ~ "inp.w;");
					mixin("ret.x = x" ~ op ~ "inp.x;");
					mixin("ret.y = y" ~ op ~ "inp.y;");
					mixin("ret.z = z" ~ op ~ "inp.z;");
					return ret;
				}

			}
			template opBinary(string op : "*")
			{
				const Vector!(qt, 3) opBinary(Vector!(qt, 3) inp)
				{
					Vector!(qt, 3) ret;
					qt ww = w ^^ 2;
					qt w2 = w * 2;
					qt wx2 = w2 * x;
					qt wy2 = w2 * y;
					qt wz2 = w2 * z;
					qt xx = x ^^ 2;
					qt x2 = x * 2;
					qt xy2 = x2 * y;
					qt xz2 = x2 * z;
					qt yy = y ^^ 2;
					qt yz2 = 2 * y * z;
					qt zz = z * z;
					ret.vector = [ww * inp.x + wy2 * inp.z - wz2 * inp.y + xx * inp.x + xy2 * inp.y + xz2 * inp.z - zz * inp.x - yy * inp.x, xy2 * inp.x + yy * inp.y + yz2 * inp.z + wz2 * inp.x - zz * inp.y + ww * inp.y - wx2 * inp.z - xx * inp.y, xz2 * inp.x + yz2 * inp.y + zz * inp.z - wy2 * inp.x - yy * inp.z + wx2 * inp.y - xx * inp.z + ww * inp.z];
					return ret;
				}

			}
			template opBinary(string op : "*")
			{
				const Quaternion opBinary(qt inp)
				{
					return Quaternion(w * inp, x * inp, y * inp, z * inp);
				}

			}
			template opOpAssign(string op : "*")
			{
				void opOpAssign(Quaternion inp)
				{
					qt w2 = -x * inp.x - y * inp.y - z * inp.z + w * inp.w;
					qt x2 = x * inp.w + y * inp.z - z * inp.y + w * inp.x;
					qt y2 = -x * inp.z + y * inp.w + z * inp.x + w * inp.y;
					qt z2 = x * inp.y - y * inp.x + z * inp.w + w * inp.z;
					w = w2;
					x = x2;
					y = y2;
					z = z2;
				}

			}
			template opOpAssign(string op) if (op == "+" || op == "-")
			{
				void opOpAssign(Quaternion inp)
				{
					mixin("w = w" ~ op ~ "inp.w;");
					mixin("x = x" ~ op ~ "inp.x;");
					mixin("y = y" ~ op ~ "inp.y;");
					mixin("z = z" ~ op ~ "inp.z;");
				}

			}
			template opOpAssign(string op : "*")
			{
				void opOpAssign(qt inp)
				{
					quaternion[0] *= inp;
					quaternion[1] *= inp;
					quaternion[2] *= inp;
					quaternion[3] *= inp;
				}

			}
			const const int opCmp(ref const Quaternion qua)
			{
				foreach (i, a; quaternion)
				{
					if (a < qua.quaternion[i])
					{
						return -1;
					}
					else
						if (a > qua.quaternion[i])
						{
							return 1;
						}
				}
				return 0;
			}


			const bool opEquals(const Quaternion qu)
			{
				return quaternion == qu.quaternion;
			}

			template opCast(T : bool)
			{
				const bool opCast()
				{
					return isFinite;
				}

			}
		}
	}
}
alias Quaternion!float quat;
