// D import file generated from 'gl3n/math.d'
module gl3n.math;
public 
{
	import std.math : PI, sin, cos, tan, asin, acos, atan, atan2, sinh, cosh, tanh, asinh, acosh, atanh, pow, exp, log, exp2, log2, sqrt, floor, trunc, round, ceil, modf;
	alias round roundEven;
	alias floor fract;
	import std.algorithm : min, max;
}
private 
{
	import std.conv : to;
	import std.algorithm : all;
	import std.array : zip;
	import std.traits : CommonType;
	import std.range : ElementType;
	import smath = std.math;
	import gl3n.util : is_vector, is_quaternion, is_matrix;
	version (unittest)
	{
		import gl3n.linalg : vec2, vec2i, vec3, vec3i, quat;
	}
}
public enum real PI_180 = PI / 180;

public enum real _180_PI = 180 / PI;

template mod(T)
{
	T mod(T x, T y)
	{
		return x - y * floor(x / y);
	}

}
nothrow pure @safe 
{
	extern (C) float fmodf(float x, float y);

	template abs(T) if (!is_vector!T && !is_quaternion!T && !is_matrix!T)
	{
		T abs(T t)
		{
			return smath.abs(t);
		}

	}
	template abs(T) if (is_vector!T)
	{
		T abs(T vec)
		{
			T ret;
			foreach (i, element; vec.vector)
			{
				ret.vector[i] = abs(element);
			}
			return ret;
		}

	}
	template abs(T) if (is_quaternion!T)
	{
		T abs(T quat)
		{
			T ret;
			ret.quaternion[0] = abs(quat.quaternion[0]);
			ret.quaternion[1] = abs(quat.quaternion[1]);
			ret.quaternion[2] = abs(quat.quaternion[2]);
			ret.quaternion[3] = abs(quat.quaternion[3]);
			return ret;
		}

	}
	real inversesqrt(real x);
	template sign(T)
	{
		float sign(T x)
		{
			if (x > 0)
			{
				return 1.00000F;
			}
			else
				if (x == 0)
				{
					return 0.00000F;
				}
				else
				{
					return -1.00000F;
				}
		}

	}
	template almost_equal(T, S) if (!is_vector!T && !is_quaternion!T)
	{
		bool almost_equal(T a, S b, float epsilon = 1e-06F)
		{
			if (abs(a - b) <= epsilon)
			{
				return true;
			}
			return abs(a - b) <= epsilon * abs(b);
		}

	}
	template almost_equal(T, S) if (is_vector!T && is_vector!S && T.dimension == S.dimension)
	{
		bool almost_equal(T a, S b, float epsilon = 1e-06F)
		{
			foreach (i; 0 .. T.dimension)
			{
				if (!almost_equal(a.vector[i], b.vector[i], epsilon))
				{
					return false;
				}
			}
			return true;
		}

	}
	template almost_equal(T) if (is_quaternion!T)
	{
		bool almost_equal(T a, T b, float epsilon = 1e-06F)
		{
			foreach (i; 0 .. 4)
			{
				if (!almost_equal(a.quaternion[i], b.quaternion[i], epsilon))
				{
					return false;
				}
			}
			return true;
		}

	}
	real radians(real degrees);
	template cradians(real degrees)
	{
		real cradians()
		{
			return radians(degrees);
		}

	}
	real degrees(real radians);
	template cdegrees(real radians)
	{
		real cdegrees()
		{
			return degrees(radians);
		}

	}
	template clamp(T1, T2, T3)
	{
		CommonType!(T1, T2, T3) clamp(T1 x, T2 min_val, T3 max_val)
		{
			return min(max(x, min_val), max_val);
		}

	}
	template step(T1, T2)
	{
		float step(T1 edge, T2 x)
		{
			return x < edge ? 0.00000F : 1.00000F;
		}

	}
	template smoothstep(T1, T2, T3)
	{
		CommonType!(T1, T2, T3) smoothstep(T1 edge0, T2 edge1, T3 x)
		{
			auto t = clamp((x - edge0) / (edge1 - edge0), 0, 1);
			return t * t * (3 - 2 * t);
		}

	}
}
