// D import file generated from 'gl3n/plane.d'
module gl3n.plane;
private 
{
	import gl3n.linalg : Vector, dot, vec3;
	import gl3n.math : almost_equal;
	import std.traits : isFloatingPoint;
}
template PlaneT(type = float) if (isFloatingPoint!type)
{
	struct PlaneT
	{
		alias type pt;
		alias Vector!(pt, 3) vec3;
		union
		{
			struct
			{
				pt a;
				pt b;
				pt c;
			}
			vec3 normal;
		}
		pt d;
		nothrow pure @safe 
		{
			this(pt a, pt b, pt c, pt d)
			{
				this.a = a;
				this.b = b;
				this.c = c;
				this.d = d;
			}

			this(vec3 normal, pt d)
			{
				this.normal = normal;
				this.d = d;
			}

			void normalize()
			{
				pt det = 1.00000 / normal.length;
				normal *= det;
				d *= det;
			}

			@property const PlaneT normalized()
			{
				PlaneT ret = PlaneT(a, b, c, d);
				ret.normalize();
				return ret;
			}


			const pt distance(vec3 point)
			{
				return dot(point, normal) + d;
			}

			const pt ndistance(vec3 point)
			{
				return (dot(point, normal) + d) / normal.length;
			}

			const bool opEquals(PlaneT other)
			{
				return other.normal == normal && other.d == d;
			}

		}
	}
}
alias PlaneT!float Plane;
